unit BasicLameTagReader;

interface

// TBasicLameTagReader reads the LAME tag from an mp3 file and
// reports its fields as basic (in most cases numeric) values.

// Copyright (c) 2003 phwip

{ TODO -cOptimise :
When a property of TBasicLameTagReader is read then a calculation
takes place to convert the relevant information from FTag.  If the
same property is read again then the same calculation takes place.
In some cases the caller reads the same property multiple times.
Should the return values be cached in local fields when first calculated
and these cached values used next time? Check out any speed/memory
implications. }

uses
	Classes,
	PresetGuesser,
	SysUtils;

const
	BitsPerByte = 8;

type
	TLameTag = packed record
		Quality: Byte;
		Encoder: array[0..3] of Char;
		VersionString: array[0..4] of Char;
		TagRevision_EncodingMethod: Byte;
		Lowpass: Byte;
		ReplayGain: packed record
			PeakSignalAmplitude: record
				case Boolean of
					True: (FloatingPoint: Single);
					False: (FixedPoint: array[0..3] of Byte);
			end;
			Track: array[0..1] of Byte;
			Album: array[0..1] of Byte;
		end;
		EncodingFlags_ATHType: Byte;
		Bitrate: Byte;
		EncoderDelays: array[0..2] of Byte;
		MiscInfo: Byte;
		MP3Gain: Byte;
		Surround_Preset: array[0..1] of Byte;
		MusicLength: array[0..3] of Byte;
		MusicCRC: array[0..1] of Byte;
		InfoTagCRC: array[0..1] of Byte;
	end;

	TReplayGainAdjustmentFields = record
		NameCode: Byte;
		Originator: Byte;
		Negative: Boolean;
		AdjustmentTimesTen: Word;
	end;

	TBasicReplayGain = record
		PeakSignalAmplitude: Single;
		Track: TReplayGainAdjustmentFields;
		Album: TReplayGainAdjustmentFields;
	end;

	TOldLameHeader = packed record
		UnusedByte: Byte;
		Encoder: array[0..3] of Char;
		VersionString: array[0..15] of Char;
	end;

	TBasicLameTagReader = class(TObject)
	private
		FDonePreset: Boolean;
		FDonePresetGuess: Boolean;
		FDoneReplayGain: Boolean;
		FOldLameHeader: TOldLameHeader;
		FPreset: Word;
		FPresetGuess: TLamePreset;
		FReplayGain: TBasicReplayGain;
		FTag: TLameTag;
		function ByteArrayToInteger(ABytes: array of Byte; const ABitsUsedPerByte:
			Byte = BitsPerByte): LongWord;
		function GetATHType: Byte;
		function GetEncoderDelay: Word;
		function GetEncodingMethod: Byte;
		function GetInfoTagCRC: Word;
		function GetLowPass: Word;
		function GetMP3GainStepChange: Shortint;
		function GetMusicCRC: Word;
		function GetMusicLength: LongWord;
		function GetNoGapContinuation: Boolean;
		function GetNoGapContinued: Boolean;
		function GetNoiseShaping: Byte;
		function GetNssafejoint: Boolean;
		function GetNspsytune: Boolean;
		function GetPaddedAtEnd: Word;
		function GetPreset: Word;
		function GetPresetGuess: TLamePreset;
		function GetReplayGain: TBasicReplayGain;
		function GetSourceSampleFrequency: Byte;
		function GetStereoMode: Byte;
		function GetSurroundInfo: Byte;
		function GetTagRevision: Byte;
		function GetUnwiseSettings: Boolean;
		function GetVersionString: String;
		function GetVersionStringNonLameTag: String;
		function SkipID3v2Tag(AFile: TFileStream): Boolean;
	public
		constructor Create(const AFileName: String);
		property ATHType: Byte read GetATHType;
		property Bitrate: Byte read FTag.Bitrate;
		property EncoderDelay: Word read GetEncoderDelay;
		property EncodingMethod: Byte read GetEncodingMethod;
		property InfoTagCRC: Word read GetInfoTagCRC;
		property Lowpass: Word read GetLowPass;
		property MP3GainStepChange: Shortint read GetMP3GainStepChange;
		property MusicCRC: Word read GetMusicCRC;
		property MusicLength: LongWord read GetMusicLength;
		property NoGapContinuation: Boolean read GetNoGapContinuation;
		property NoGapContinued: Boolean read GetNoGapContinued;
		property NoiseShaping: Byte read GetNoiseShaping;
		property Nssafejoint: Boolean read GetNssafejoint;
		property Nspsytune: Boolean read GetNspsytune;
		property PaddedAtEnd: Word read GetPaddedAtEnd;
		property Preset: Word read GetPreset;
		property PresetGuess: TLamePreset read GetPresetGuess;
		property Quality: Byte read FTag.Quality;
		property ReplayGain: TBasicReplayGain read GetReplayGain;
		property SourceSampleFrequency: Byte read GetSourceSampleFrequency;
		property StereoMode: Byte read GetStereoMode;
		property SurroundInfo: Byte read GetSurroundInfo;
		property TagRevision: Byte read GetTagRevision;
		property UnwiseSettings: Boolean read GetUnwiseSettings;
		property VersionString: String read GetVersionString;
		property VersionStringNonLameTag: String read GetVersionStringNonLameTag;
	end;

	ELameTagReaderExpected = class(Exception);

implementation

uses
	JclLogic;

const
	InfoOffset1 = $0D;
	InfoOffset2 = $15;
	InfoOffset3  = $24;
	LAMETagOffset = $77;

type
	TID3v2Header = packed record
		Identifier: array[0..2] of Char;
		Version: Word;
		Flags: Byte;
		Size: array[0..3] of Byte;
	end;

	TStartOfFile = packed record
		Misc1: array[0..InfoOffset1 - 1] of Byte;
		Info1: array[0..3] of Char;
		Misc2: array[InfoOffset1 + 4..InfoOffset2 - 1] of Byte;
		Info2: array[0..3] of Char;
		Misc3: array[InfoOffset2 + 4..InfoOffset3 - 1] of Byte;
		Info3: array[0..3] of Char;
	end;

{ TBasicLameTagReader }

function TBasicLameTagReader.ByteArrayToInteger(ABytes: array of Byte; const
	ABitsUsedPerByte: Byte): LongWord;
var
	LoopCounter: Integer;
begin
	Result := 0;
	for LoopCounter := Low(ABytes) to High(ABytes) do
		Result := Result or ABytes[LoopCounter] shl ((High(ABytes) - LoopCounter) *
			ABitsUsedPerByte);
end;

constructor TBasicLameTagReader.Create(const AFileName: String);
var
	Mp3File: TFileStream;
	StartOfFile: TStartOfFile;
begin
	inherited Create;
	FDonePreset := False;
	FDonePresetGuess := False;
	FDoneReplayGain := False;
	Mp3File := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyNone);
	try
		SkipID3v2Tag(Mp3File);
		Mp3File.Read(StartOfFile, SizeOf(StartOfFile));
		Mp3File.Seek(-SizeOf(StartOfFile), soFromCurrent);
		if (StartOfFile.Info1 = 'Xing') or (StartOfFile.Info1 = 'Info') then
			Mp3File.Seek(InfoOffset1, soFromCurrent)
		else if (StartOfFile.Info2 = 'Xing') or (StartOfFile.Info2 = 'Info') then
			Mp3File.Seek(InfoOffset2, soFromCurrent)
		else if (StartOfFile.Info3 = 'Xing') or (StartOfFile.Info3 = 'Info') then
			Mp3File.Seek(InfoOffset3, soFromCurrent)
		else
			raise ELameTagReaderExpected.Create('LAME tag not found.');
		Mp3File.Seek(LAMETagOffset, soFromCurrent);
		if not (Mp3File.Read(FTag, SizeOf(FTag)) = SizeOf(FTag)) then
			raise Exception.Create('Unexpected End Of File.');
		Mp3File.Seek(-SizeOf(FTag), soFromCurrent);
		Mp3File.Read(FOldLameHeader, SizeOf(FOldLameHeader));
{ TODO : How about other encoders that use the lame tag? }
		if FTag.Encoder <> 'LAME' then
			raise ELameTagReaderExpected.Create('LAME tag not found.');
	finally
		FreeAndNil(Mp3File);
	end;
end;

function TBasicLameTagReader.GetATHType: Byte;
begin
	Result := FTag.EncodingFlags_ATHType and $0F;
end;

function TBasicLameTagReader.GetEncoderDelay: Word;
begin
	Result := (FTag.EncoderDelays[0] shl 4) or ((FTag.EncoderDelays[1] and $F0)
		shr 4);
end;

function TBasicLameTagReader.GetEncodingMethod: Byte;
begin
	Result := FTag.TagRevision_EncodingMethod and $0F;
end;

function TBasicLameTagReader.GetInfoTagCRC: Word;
begin
	Result := ByteArrayToInteger(FTag.InfoTagCRC);
end;

function TBasicLameTagReader.GetLowPass: Word;
begin
	Result := FTag.LowPass * 100;
end;

function TBasicLameTagReader.GetMP3GainStepChange: Shortint;
begin
	Result := FTag.MP3Gain and $7F;
	if TestBit(FTag.MP3Gain, 7) then
		Result := -Result;
end;

function TBasicLameTagReader.GetMusicCRC: Word;
begin
	Result := ByteArrayToInteger(FTag.MusicCRC);
end;

function TBasicLameTagReader.GetMusicLength: LongWord;
begin
	Result := ByteArrayToInteger(FTag.MusicLength);
end;

function TBasicLameTagReader.GetNoGapContinuation: Boolean;
begin
	Result := TestBit(FTag.EncodingFlags_ATHType, 7);
end;

function TBasicLameTagReader.GetNoGapContinued: Boolean;
begin
	Result := TestBit(FTag.EncodingFlags_ATHType, 6);
end;

function TBasicLameTagReader.GetNoiseShaping: Byte;
begin
	Result := FTag.MiscInfo and $03;
end;

function TBasicLameTagReader.GetNspsytune: Boolean;
begin
	Result := TestBit(FTag.EncodingFlags_ATHType, 4);
end;

function TBasicLameTagReader.GetNssafejoint: Boolean;
begin
	Result := TestBit(FTag.EncodingFlags_ATHType, 5);
end;

function TBasicLameTagReader.GetPaddedAtEnd: Word;
begin
	Result := ((FTag.EncoderDelays[1] and $0F) shl 8) or FTag.EncoderDelays[2];
end;

function TBasicLameTagReader.GetPreset: Word;
begin
	if not FDonePreset then
	begin
		FDonePreset := True;
		FPreset := ByteArrayToInteger(FTag.Surround_Preset) and $07FF;
	end;
	Result := FPreset;
end;

function TBasicLameTagReader.GetPresetGuess: TLamePreset;
begin
	if not FDonePresetGuess then
	begin
		FDonePresetGuess := True;
		FPresetGuess := PresetGuesser.GuessPreset(VersionString, Bitrate, Quality,
			EncodingMethod, NoiseShaping, StereoMode, ATHType, FTag.LowPass);
	end;
	Result := FPresetGuess;
end;

function TBasicLameTagReader.GetReplayGain: TBasicReplayGain;

	procedure SetAdjustmentFields(var AFlds: TReplayGainAdjustmentFields;
		ATagInfo: array of Byte);
	begin
		AFlds.NameCode := (ATagInfo[0] and $E0) shr 5;
		AFlds.Originator := (ATagInfo[0] and $1C) shr 2;
		AFlds.Negative := TestBit(ATagInfo[0], 1);
		AFlds.AdjustmentTimesTen := ((ATagInfo[0] and $01) shl 8) or ATagInfo[1];
	end;

const
	FirstLameFloatingPointTrackPeak = '3.94b';
begin
	if not FDoneReplayGain then
	begin
		FDoneReplayGain := True;
		if VersionString < FirstLameFloatingPointTrackPeak then
			FReplayGain.PeakSignalAmplitude := FTag.ReplayGain.PeakSignalAmplitude.
				FloatingPoint
		else
			FReplayGain.PeakSignalAmplitude := ByteArrayToInteger(FTag.ReplayGain.
				PeakSignalAmplitude.FixedPoint) / $800000;
		SetAdjustmentFields(FReplayGain.Track, FTag.ReplayGain.Track);
		SetAdjustmentFields(FReplayGain.Album, FTag.ReplayGain.Album);
	end;
	Result := FReplayGain;
end;

function TBasicLameTagReader.GetSourceSampleFrequency: Byte;
begin
	Result := (FTag.MiscInfo and $C0) shr 6;
end;

function TBasicLameTagReader.GetStereoMode: Byte;
begin
	Result := (FTag.MiscInfo and $1C) shr 2;
end;

function TBasicLameTagReader.GetSurroundInfo: Byte;
begin
	Result := (FTag.Surround_Preset[0] and $38) shr 3;
end;

function TBasicLameTagReader.GetTagRevision: Byte;
begin
	Result := (FTag.TagRevision_EncodingMethod and $F0) shr 4;
end;

function TBasicLameTagReader.GetUnwiseSettings: Boolean;
begin
	Result := TestBit(FTag.MiscInfo, 5);
end;

function TBasicLameTagReader.GetVersionString: String;
begin
	Result := Format('%.5s', [FTag.VersionString]);
end;

function TBasicLameTagReader.GetVersionStringNonLameTag: String;
{
	In versions of LAME before the LAME tag was introduced, a 16 byte version
	string was written into the LAME header, starting at the same position as
	the current version string.
}
begin
	Result := Format('%.16s', [FOldLameHeader.VersionString]);
end;

function TBasicLameTagReader.SkipID3v2Tag(AFile: TFileStream): Boolean;
var
	Header: TID3v2Header;
	TagSize: LongWord;
begin
	if not (AFile.Read(Header, SizeOf(Header)) = SizeOf(Header)) then
		raise Exception.Create('Unexpected End Of File.');
	Result := Header.Identifier = 'ID3';
	if Result then
	begin
		TagSize := ByteArrayToInteger(Header.Size, 7);
		if TestBit(Header.Flags, 4) then	// ID3v2 flag has ten byte footer
			Inc(TagSize, $0A);
		AFile.Seek(TagSize, soFromCurrent);
	end
	else
		AFile.Seek(-SizeOf(Header), soFromCurrent);
end;

end.
