unit DescriptiveLameTagReader;

// TDescriptiveLameTagReader uses TBasicLameTagReader to read the LAME tag
// from an mp3 file but reports its fields as descriptive (in most cases
// textual) values.

// Copyright (c) 2003 phwip

interface

uses
	BasicLameTagReader;

type
	TDescriptiveReplayGain = record
		PeakSignalAmplitude: String;
		Track: String;
		Album: String;
	end;

	TUsePresetGuess = (upgNotNeeded, upgUseGuess, upgUnableToGuess);

	TDescriptiveLameTagReader = class(TObject)
	private
		FBasicReader: TBasicLameTagReader;
		FDonePresetRelatedValues: Boolean;
		FPreset: String;
		FPresetGuess: String;
		FUsePresetGuess: TUsePresetGuess;
		function BoolToNoYes(const ABool: Boolean): String;
		function BoolToPlusMinus(const ABool: Boolean): String;
		procedure DeterminePresetRelatedValues;
		function GetATHType: Byte;
		function GetBitrate: String;
		function GetEncoderDelay: String;
		function GetEncodingMethod: String;
		function GetInfoTagCRC: String;
		function GetLowPass: String;
		function GetMP3GainDBChange: String;
		function GetMusicCRC: String;
		function GetMusicLength: String;
		function GetNoGapContinuation: String;
		function GetNoGapContinued: String;
		function GetNoiseShaping: Byte;
		function GetNspsytune: String;
		function GetNssafejoint: String;
		function GetPaddedAtEnd: String;
		function GetPreset: String;
		function GetPresetGuess: String;
		function GetQuality: String;
		function GetReplayGain: TDescriptiveReplayGain;
		function GetSourceSampleFrequency: String;
		function GetStereoMode: String;
		function GetSurroundInfo: String;
		function GetTagRevision: Byte;
		function GetUnwiseSettings: String;
		function GetVersionString: String;
		function GetVersionStringNonLameTag: String;
		function ThousandsSeparated(const I: Integer): String;
    function GetUsePresetGuess: TUsePresetGuess;
	public
		constructor Create(const AFileName: String);
		destructor Destroy; override;
		property ATHType: Byte read GetATHType;
		property Bitrate: String read GetBitrate;
		property EncoderDelay: String read GetEncoderDelay;
		property EncodingMethod: String read GetEncodingMethod;
		property InfoTagCRC: String read GetInfoTagCRC;
		property Lowpass: String read GetLowPass;
		property MP3GainDBChange: String read GetMP3GainDBChange;
		property MusicCRC: String read GetMusicCRC;
		property MusicLength: String read GetMusicLength;
		property NoGapContinuation: String read GetNoGapContinuation;
		property NoGapContinued: String read GetNoGapContinued;
		property NoiseShaping: Byte read GetNoiseShaping;
		property Nssafejoint: String read GetNssafejoint;
		property Nspsytune: String read GetNspsytune;
		property PaddedAtEnd: String read GetPaddedAtEnd;
		property Preset: String read GetPreset;
		property PresetGuess: String read GetPresetGuess;
		property Quality: String read GetQuality;
		property ReplayGain: TDescriptiveReplayGain read GetReplayGain;
		property SourceSampleFrequency: String read GetSourceSampleFrequency;
		property StereoMode: String read GetStereoMode;
		property SurroundInfo: String read GetSurroundInfo;
		property TagRevision: Byte read GetTagRevision;
		property UnwiseSettings: String read GetUnwiseSettings;
		property UsePresetGuess: TUsePresetGuess read GetUsePresetGuess;
		property VersionString: String read GetVersionString;
		property VersionStringNonLameTag: String read GetVersionStringNonLameTag;
	end;

implementation

uses
	PresetGuesser,
	SysUtils;

{ TDescriptiveLameTagReader }

function TDescriptiveLameTagReader.BoolToNoYes(const ABool: Boolean): String;
const
	NoYes: array[Boolean] of String = ('no', 'yes');
begin
	Result := NoYes[ABool];
end;

function TDescriptiveLameTagReader.BoolToPlusMinus(const ABool: Boolean):
	String;
const
	PlusMinus: array[Boolean] of Char = ('+', '-');
begin
	Result := PlusMinus[ABool];
end;

constructor TDescriptiveLameTagReader.Create(const AFileName: String);
begin
	inherited Create;
	FBasicReader := TBasicLameTagReader.Create(AFileName);
	FDonePresetRelatedValues := False;
end;

destructor TDescriptiveLameTagReader.Destroy;
begin
	FreeAndNil(FBasicReader);
	inherited;
end;

procedure TDescriptiveLameTagReader.DeterminePresetRelatedValues;

	function DeterminePreset(var AUsePresetGuess: TUsePresetGuess): String;
	begin
		AUsePresetGuess := upgNotNeeded;
		case FBasicReader.Preset of
		0:
		begin
			Result := '<not recorded>';
			AUsePresetGuess := upgUseGuess;
		end;
		8..320:
		begin
			Result := IntToStr(FBasicReader.Preset);
			if FBasicReader.EncodingMethod = 1 then
				Result := 'cbr ' + Result;
			AUsePresetGuess := upgUseGuess;
		end;
		410:
			Result := 'V9';
		420:
			Result := 'V8';
		430:
			Result := 'V7';
		440:
			Result := 'V6';
		450:
			Result := 'V5';
		460:
			Result := 'V4: preset medium';
		470:
			Result := 'V3';
		480:
			Result := 'V2: preset standard';
		490:
			Result := 'V1';
		500:
			Result := 'V0: preset extreme';
		1000:
			Result := 'r3mix';
		1001:
			Result := '[alt-]preset standard';
		1002:
			Result := '[alt-]preset extreme';
		1003:
			Result := '[alt-]preset insane';
		1004:
			Result := '[alt-]preset fast standard';
		1005:
			Result := '[alt-]preset fast extreme';
		1006:
			Result := '[alt-]preset medium';
		1007:
			Result := '[alt-]preset fast medium';
		1010:
			Result := 'preset portable';	// only used in alpha versions (3.94 alpha 14 and below)
		1015:
			Result := 'preset radio';	// only used in alpha versions (3.94 alpha 14 and below)
		else
		begin
			Result := Format('<unrecognised value %d>', [FBasicReader.Preset]);
			AUsePresetGuess := upgUseGuess;
		end
		end;
		{
			Delphi doesn't allow sets to contain values larger than 255, so subtract
			400 and compare to 10, 20... to test for values 410, 420...
		}
		if ((FBasicReader.Preset - 400) in [10, 20, 30, 40, 50, 60, 70, 80, 90,
				100]) and (FBasicReader.EncodingMethod = 4) then
			Result := Result + ' (fast mode)';
	end;

	function DeterminePresetGuess(var AUsePresetGuess: TUsePresetGuess): String;
	begin
		case FBasicReader.PresetGuess of
		lpInsane:
			Result := '[alt-]preset insane';
		lpExtreme:
			Result := '[alt-]preset extreme';
		lpFastExtreme:
			Result := '[alt-]preset fast extreme';
		lpStandard:
			Result := '[alt-]preset standard';
		lpFastStandard:
			Result := '[alt-]preset fast standard';
		lpR3mix:
			Result := 'r3mix';
		lpStudio:
			Result := 'preset studio';
		lpCD:
			Result := 'preset cd';
		lpHifi:
			Result := 'preset hifi';
		lpTape:
			Result := 'preset tape';
		lpRadio:
			Result := 'preset radio';
		lpFM:
			Result := 'preset fm';
		lpTapeRadioFM:
			Result := 'preset tape OR preset radio OR preset fm';
		lpVoice:
			Result := 'preset voice';
		lpMWUS:
			Result := 'preset mw-us';
		lpMWEU:
			Result := 'preset phon+ OR preset lw OR preset mw-eu OR preset sw';
		lpPhone:
			Result := 'preset phone';
		else
		begin
			Result := '';
			{
				Only display "unable to guess" if the preset was not recorded (ie. the
				field value is zero.
			}
			if FBasicReader.Preset = 0 then
				AUsePresetGuess := upgUnableToGuess
			else
				AUsePresetGuess := upgNotNeeded;
		end;
		end;
	end;

begin
	if FDonePresetRelatedValues then
		Exit;
	FDonePresetRelatedValues := True;
	FPreset := DeterminePreset(FUsePresetGuess);
	if FUsePresetGuess = upgNotNeeded then
		FPresetGuess := ''
	else
		FPresetGuess := DeterminePresetGuess(FUsePresetGuess);
end;

function TDescriptiveLameTagReader.GetATHType: Byte;
begin
	Result := FBasicReader.ATHType;
end;

function TDescriptiveLameTagReader.GetBitrate: String;
begin
	if FBasicReader.Bitrate = $00 then
		Result := '<unknown>'
	else
	begin
		Result := '';
		if FBasicReader.EncodingMethod in [3..6] then			// vbr
			Result := 'minimal (-b) bitrate ';
		Result := Result + IntToStr(FBasicReader.Bitrate);
		if FBasicReader.Bitrate = $FF then
			Result := Result + ' or higher';
	end;
end;

function TDescriptiveLameTagReader.GetEncoderDelay: String;
begin
	Result := Format('%s samples', [ThousandsSeparated(FBasicReader.
		EncoderDelay)]);
end;

function TDescriptiveLameTagReader.GetEncodingMethod: String;
begin
	case FBasicReader.EncodingMethod of
	0:
		Result := '<unknown>';
	1:
		Result := 'cbr';
	2:
		Result := 'abr';
	3:
		Result := 'vbr old / vbr rh';
	4:
		Result := 'vbr new / vbr mtrh';
	5:
		Result := 'vbr mt';
	8:
		Result := 'cbr 2 pass';
	9:
		Result := 'abr 2 pass';
	else
		Result := Format('<not recognised: internal value %d>', [FBasicReader.
			EncodingMethod]);
	end;
end;

function TDescriptiveLameTagReader.GetInfoTagCRC: String;
begin
	Result := Format('%.4x', [FBasicReader.InfoTagCRC]);
end;

function TDescriptiveLameTagReader.GetLowPass: String;
begin
	if FBasicReader.Lowpass = 0 then
		Result := '<unknown>'
	else
		Result := Format('%sHz', [ThousandsSeparated(FBasicReader.Lowpass)]);
end;

function TDescriptiveLameTagReader.GetMP3GainDBChange: String;
var
	BRMP3Gain: Shortint;
begin
	BRMP3Gain := FBasicReader.MP3GainStepChange;
	if BRMP3Gain = 0 then
		Result := '<none>'
	else
		Result := Format('%s%.1u.%.1udB', [BoolToPlusMinus(BRMP3Gain < 0), (Abs(
			BRMP3Gain) * 15) div 10, (Abs(BRMP3Gain) * 15) mod 10]);
end;

function TDescriptiveLameTagReader.GetMusicCRC: String;
begin
	Result := Format('%.4x', [FBasicReader.MusicCRC]);
end;

function TDescriptiveLameTagReader.GetMusicLength: String;
begin
	if FBasicReader.MusicLength = 0 then
		Result := '<unknown or larger than 4GB>'
	else
		Result := Format('%0.n bytes', [0.0 + FBasicReader.MusicLength]);
end;

function TDescriptiveLameTagReader.GetNoGapContinuation: String;
begin
	Result := BoolToNoYes(FBasicReader.NoGapContinuation);
end;

function TDescriptiveLameTagReader.GetNoGapContinued: String;
begin
	Result := BoolToNoYes(FBasicReader.NoGapContinued);
end;

function TDescriptiveLameTagReader.GetNoiseShaping: Byte;
begin
	Result := FBasicReader.NoiseShaping;
end;

function TDescriptiveLameTagReader.GetNspsytune: String;
begin
	Result := BoolToNoYes(FBasicReader.Nspsytune);
end;

function TDescriptiveLameTagReader.GetNssafejoint: String;
begin
	Result := BoolToNoYes(FBasicReader.Nssafejoint);
end;

function TDescriptiveLameTagReader.GetPaddedAtEnd: String;
begin
	Result := Format('%s samples', [ThousandsSeparated(FBasicReader.
		PaddedAtEnd)]);
end;

function TDescriptiveLameTagReader.GetPreset: String;
begin
	DeterminePresetRelatedValues;
	Result := FPreset;
end;

function TDescriptiveLameTagReader.GetPresetGuess: String;
begin
	DeterminePresetRelatedValues;
	Result := FPresetGuess;
end;

function TDescriptiveLameTagReader.GetQuality: String;
var
	BRQuality: Byte;
begin
	BRQuality := FBasicReader.Quality;
	if BRQuality in [1..100] then
		Result := Format('%d (V%d and q%d)', [BRQuality, (100 - BRQuality) div 10,
			(100 - BRQuality) mod 10])
	else
		Result := Format('<unexpected value: %d>', [BRQuality]);
end;

function TDescriptiveLameTagReader.GetReplayGain: TDescriptiveReplayGain;

	function SetBy(const AOriginator: Byte): String;
	const
		SetBy: array[1..4] of String =
		(
			'set by artist',
			'set by user',
			'determined automatically',
			'set by simple RMS average'
		);
	begin
		if (AOriginator < Low(SetBy)) or (AOriginator > High(SetBy)) then
			Result := '<set by unknown>'
		else
			Result := SetBy[AOriginator];
	end;

	function AdjString(AFlds: TReplayGainAdjustmentFields; const
		AExpectedNameCode: Byte): String;
	begin
		if (AFlds.NameCode = AExpectedNameCode) and (AFlds.Originator <> 0) then
			Result := Format('%s%.1u.%.1udB (%s)', [BoolToPlusMinus(AFlds.Negative),
				AFlds.AdjustmentTimesTen div 10, AFlds.AdjustmentTimesTen mod 10,
				SetBy(AFlds.Originator)])
		else
			Result := '<not stored>';
	end;

begin
	if FBasicReader.ReplayGain.PeakSignalAmplitude = 0 then
		Result.PeakSignalAmplitude := '<not stored>'
	else
		Result.PeakSignalAmplitude := FloatToStrF(FBasicReader.ReplayGain.
			PeakSignalAmplitude, ffFixed, 8, 6);
	Result.Track := AdjString(FBasicReader.ReplayGain.Track, 1);
	Result.Album := AdjString(FBasicReader.ReplayGain.Album, 2);
end;

function TDescriptiveLameTagReader.GetSourceSampleFrequency: String;
const
	DescSF: array[0..3] of String =
	(
		'32kHz or smaller',
		'44.1kHz',
		'48kHz',
		'higher than 48kHz'
	);
begin
	Result := DescSF[FBasicReader.SourceSampleFrequency];
end;

function TDescriptiveLameTagReader.GetStereoMode: String;
const
	DescSM: array[0..7] of String =
	(
		'mono',
		'stereo',
		'dual',
		'joint',
		'force',
		'auto',
		'intensity',
		'<undefined>'
	);
begin
	Result := DescSM[FBasicReader.StereoMode];
end;

function TDescriptiveLameTagReader.GetSurroundInfo: String;
begin
	case FBasicReader.SurroundInfo of
	0:
		Result := 'none';
	1:
		Result := 'DPL Encoding';
	2:
		Result := 'DPL2 Encoding';
	3:
		Result := 'Ambisonic Encoding';
	else
		Result := Format('<not recognised: internal value %d>', [FBasicReader.
			SurroundInfo]);
	end;
end;

function TDescriptiveLameTagReader.GetTagRevision: Byte;
begin
	Result := FBasicReader.TagRevision;
end;

function TDescriptiveLameTagReader.GetUnwiseSettings: String;
begin
	Result := BoolToNoYes(FBasicReader.UnwiseSettings);
end;

function TDescriptiveLameTagReader.GetUsePresetGuess: TUsePresetGuess;
begin
	DeterminePresetRelatedValues;
	Result := FUsePresetGuess;
end;

function TDescriptiveLameTagReader.GetVersionString: String;
begin
	Result := FBasicReader.VersionString;
end;

function TDescriptiveLameTagReader.GetVersionStringNonLameTag: String;
begin
	Result := FBasicReader.VersionStringNonLameTag;
end;

function TDescriptiveLameTagReader.ThousandsSeparated(const I: Integer):
	String;
begin
	Result := Format('%0.n', [0.0 + I]);
end;

end.
