program LameTag;

{$APPTYPE CONSOLE}

{$R 'VersionInfo.res' 'VersionInfo.rc'}

// Command line utility to display the LAME tag from an mp3 file
// Copyright (c) 2003 phwip

uses
	BasicLameTagReader in 'BasicLameTagReader.pas',
	DescriptiveLameTagReader in 'DescriptiveLameTagReader.pas',
	PresetGuesser in 'PresetGuesser.pas',
	JclFileUtils,
	SysUtils;

procedure Write(const AString: String = ''); overload;
begin
	WriteLn(AString);
end;

procedure Write(const AString: String; const Args: array of const); overload;
begin
	Write(Format(AString, Args));
end;

procedure ReportException(const E: Exception);
begin
	if E is ELameTagReaderExpected then
		Write(E.Message)
	else
		Write('ERROR: %s', [E.Message]);
end;

procedure ReadLameTag(const AFileName: String);

	procedure WarnOnUnsupportedTagRevision(const ATagRevision: Byte);
	const
		MaxSupportedTagRevision = 0;
	begin
		if ATagRevision > MaxSupportedTagRevision then
		begin
			Write;
			Write('WARNING:');
			Write('The tag revision of the LAME tag in this file is %d.',
				[ATagRevision]);
			Write('This version of LameTag.exe only supports tag revisions up');
			Write('to %d, so the values shown will probably be incorrect.', [
				MaxSupportedTagRevision]);
			Write;
		end;
	end;

const
	FirstLameWithTag = '3.90';
var
	Reader: TDescriptiveLameTagReader;
begin
	Reader := TDescriptiveLameTagReader.Create(AFileName);
	try
		if Reader.VersionString < FirstLameWithTag then
		begin
			Write;
			Write('LAME tags were introduced with LAME version %s, but ', [
				FirstLameWithTag]);
			Write('this file was encoded by an earlier version of LAME and');
			Write('the header only contains a version string.');
			Write;
			Write('Version string:     %s', [Reader.VersionStringNonLameTag]);
		end
		else
		begin
			WarnOnUnsupportedTagRevision(Reader.TagRevision);
			Write('Tag revision:       %d', [Reader.TagRevision]);
			Write('Version string:     %s', [Reader.VersionString]);
			Write('Quality:            %s', [Reader.Quality]);
			Write('Encoding method:    %s', [Reader.EncodingMethod]);
			Write('Lowpass:            %s', [Reader.LowPass]);
			Write('RG track peak:      %s', [Reader.ReplayGain.PeakSignalAmplitude]);
			Write('RG track gain:      %s', [Reader.ReplayGain.Track]);
			Write('RG album gain:      %s', [Reader.ReplayGain.Album]);
			Write('nspsytune:          %s', [Reader.Nspsytune]);
			Write('nssafejoint:        %s', [Reader.Nssafejoint]);
			Write('nogap continued:    %s', [Reader.NoGapContinued]);
			Write('nogap continuation: %s', [Reader.NoGapContinuation]);
			Write('ATH type:           %d', [Reader.ATHType]);
			Write('Bitrate:            %s', [Reader.Bitrate]);
			Write('Encoder delay:      %s', [Reader.EncoderDelay]);
			Write('Padded at end:      %s', [Reader.PaddedAtEnd]);
			Write('Noise shaping:      %d', [Reader.NoiseShaping]);
			Write('Stereo mode:        %s', [Reader.StereoMode]);
			Write('Unwise settings:    %s', [Reader.UnwiseSettings]);
			Write('Source sample freq: %s', [Reader.SourceSampleFrequency]);
			Write('MP3Gain change:     %s', [Reader.MP3GainDBChange]);
			case Reader.UsePresetGuess of
			upgUseGuess:
				Write('Preset:             %s - see preset guess at end', [Reader.
					Preset]);
			upgUnableToGuess:
				Write('Preset:             %s - unable to guess', [Reader.Preset]);
			else
				Write('Preset:             %s', [Reader.Preset]);
			end;
			Write('Surround info:      %s', [Reader.SurroundInfo]);
			Write('Music length:       %s', [Reader.MusicLength]);
			Write('Music CRC:          %s', [Reader.MusicCRC]);
			Write('Info tag CRC:       %s', [Reader.InfoTagCRC]);
			if Reader.UsePresetGuess = upgUseGuess then
			begin
				Write;
				Write('  Based on a combination of other values from the lame tag,');
				Write('  it is likely that the preset used to encode this file was:');
				Write;
				Write('Preset guess:       %s', [Reader.PresetGuess]);
			end;
		end;
	finally
		FreeAndNil(Reader);
	end;
end;

function ReadLameTags(const AFileNameMask: String; const ARecursive: Boolean):
	Boolean;
// Returns False if no files match AFileNameMask
var
	Found: Integer;
	FileSearch, SubFolderSearch: TSearchRec;
	FullPathMask: String;
	Folder: String;
	FileNameMask: String;
	FullPathFileName: String;
begin
	FullPathMask := ExpandFileName(AFileNameMask);
	if IsDirectory(FullPathMask) then
		FullPathMask := JclFileUtils.PathAddSeparator(FullPathMask) + '*';
	Folder := JclFileUtils.PathAddSeparator(ExtractFilePath(FullPathMask));
	FileNameMask := ExtractFileName(FullPathMask);
	Found := FindFirst(FullPathMask, faReadOnly + faHidden + faArchive +
		faSysFile, FileSearch);
	try
		Result := Found = 0;
		while Found = 0 do
		begin
			FullPathFileName := Folder + FileSearch.Name;
			Write(FullPathFileName);
			try
				ReadLameTag(FullPathFileName);
			except
				on E: Exception do
					ReportException(E);
			end;
			Write;
			Found := FindNext(FileSearch);
		end;
	finally
		FindClose(FileSearch);
	end;
	if ARecursive then
	begin
		Found := FindFirst(Folder + '*', faDirectory, SubFolderSearch);
		try
			while Found = 0 do
			begin
				if (SubFolderSearch.Name <> '.') and (SubFolderSearch.Name <> '..') and
						((SubFolderSearch.Attr and faDirectory) = faDirectory) then
					ReadLameTags(JclFileUtils.PathAddSeparator(Folder + SubFolderSearch.
						Name) + FileNameMask, ARecursive);
				Found := FindNext(SubFolderSearch);
			end;
		finally
			FindClose(SubFolderSearch);
		end;
	end;
end;

procedure Titles(const AExeName: String);
var
	VerInfo: TJclFileVersionInfo;
begin
	VerInfo := TJclFileVersionInfo.Create(AExeName);
	try
		Write('%s - %s', [VerInfo.ProductName, VerInfo.FileDescription]);
		Write('Copyright %s', [VerInfo.LegalCopyright]);
		Write('Release %s, compiled %s', [VerInfo.ProductVersion, VerInfo.Items.
			Values['CompileDate']]);
	finally
		FreeAndNil(VerInfo);
	end;
end;

procedure Usage;
begin
	Write('Usage: LameTag [--recursive] <filename|folder>');
	Write('       Wildcards are supported in the filename');
end;

procedure Notes;
begin
	Write('Notes:');
	Write('- The LAME version string in the LAME tag is only five characters');
	Write('  long, so version numbers such as 3.93.1 are truncated.');
	Write('- The preset field is only written by LAME version 3.93 and higher');
	Write('  and some later compiles of version 3.90.3.  For files where the');
	Write('  preset was not written by the encoder, an attempt will be made');
	Write('  to guess the preset, based on the following fields: bitrate,');
	Write('  quality, encoding method, noise shaping, stereo mode, ATH type');
	Write('  and lowpass.');
	Write('- The ReplayGain fields are only written by LAME version 3.94 and');
	Write('  higher.');
	Write('- The MP3Gain field in the LAME tag is not written by any current');
	Write('  implementations of MP3Gain.');
end;

const
	HelpStrings: array[0..5] of String =
	(
		'help',
		'-help',
		'--help',
		'?',
		'-?',
		'/?'
	);
	OptionRecursive = '--recursive';
var
	LoopCounter: Byte;
	Recursive: Boolean;
	FileNameParam: Integer;
begin
	try
		Titles(ParamStr(0));
		Write;
		Recursive := ParamStr(1) = OptionRecursive;
		if Recursive then
			FileNameParam := 2
		else
			FileNameParam := 1;
		for LoopCounter := Low(HelpStrings) to High(HelpStrings) do
			if ParamStr(1) = HelpStrings[LoopCounter] then
				Break;
		if ParamStr(1) = HelpStrings[LoopCounter] then
		begin
			Usage;
			Write;
			Notes;
		end
		else if ParamCount <> FileNameParam then
			Usage
		else if not ReadLameTags(ParamStr(FileNameParam), Recursive) then
		begin
			Usage;
			Write;
			Write('FILE NOT FOUND: ''%s''', [ParamStr(1)]);
		end;
	except
		on E: Exception do
			ReportException(E);
	end;
end.
