unit PresetGuesser;

// function GuessPreset uses values from the LAME tag to guess the preset
// that was used where this is not stored in the LAME tag.  This is the case
// for versions of LAME earlier than 3.93 (except later releases of 3.90.3).

// Copyright (c) 2003 phwip

interface

type
	TLamePreset =
	(
		lpInsane, // --(alt-)preset insane
		lpExtreme, // --(alt-)preset extreme
		lpFastExtreme, // --(alt-)preset fast extreme
		lpStandard, // --(alt-)preset standard
		lpFastStandard, // --(alt-)preset fast standard
		lpR3mix, // --r3mix
		lpStudio, // --preset studio
		lpCD, // --preset cd
		lpHifi, // --preset hifi
		lpTape, // --preset tape
		lpRadio, // --preset radio
		lpFM, // --preset fm
		lpTapeRadioFM, // --preset tape OR --preset radio OR --preset fm
									// the above three presets generate identical files for
									// versions of lame from 3.90.3 and 3.93 up
		lpVoice, // --preset voice
		lpMWUS, // --preset mw-us
		lpMWEU, // --preset phon+ OR --preset lw OR --preset mw-eu OR --preset sw
		lpPhone, // --preset phone
		lpUnknown
	);

function GuessPreset(const AVersionString: String; const ABitrate, AQuality,
	AEncodingMethod, ANoiseShaping, AStereoMode, AATHType, ALowpassDiv100: Byte):
	TLamePreset;

implementation

type

	{
		Different versions of lame are grouped together in TLameVersionGroup if
		they share the same PresetGuessTable values.
	}
	TLameVersionGroup = (lvg390_3901_392, lvg3902_391, lvg393, lvg3931_3903up,
		lvg394up);

	{
		TTagValues array is made up of:
			Bitrate (as recorded in the lame tag - may be abr or minimal bitrate)
			Quality
			EncodingMethod
			NoiseShaping
			StereoMode
			ATHType
			Lowpass divided by 100
	}
	TTagValues = array[0..6] of Byte;

	TPresetGuessRow = record
		VGs: set of TLameVersionGroup; // Version Groups
		TVs: array[0..6] of Byte;      // Tag Values
		Res: TLamePreset;              // Resulting Lame Preset
	end;

const
	PresetGuessTable: array[0..40] of TPresetGuessRow =
	(
		(
			VGs: [lvg390_3901_392]; // 3.90, 3.90.1, 3.92
			TVs: (255, 58, 1, 1, 3, 2, 205);
			Res: lpInsane
		),
		(
			VGs: [lvg3902_391, lvg3931_3903up]; // 3.90.2, 3.90.3, 3.91
			TVs: (255, 58, 1, 1, 3, 2, 206);
			Res: lpInsane
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (255, 57, 1, 1, 3, 4, 205);
			Res: lpInsane
		),
		(
			VGs: [lvg390_3901_392]; // 3.90, 3.90.1, 3.92
			TVs: (000, 78, 3, 2, 3, 2, 195);
			Res: lpExtreme
		),
		(
			VGs: [lvg3902_391]; // 3.90.2, 3.91
			TVs: (000, 78, 3, 2, 3, 2, 196);
			Res: lpExtreme
		),
		(
			VGs: [lvg3931_3903up]; // 3.90.3
			TVs: (000, 78, 3, 1, 3, 2, 196);
			Res: lpExtreme
		),
		(
			VGs: [lvg390_3901_392]; // 3.90, 3.90.1, 3.92
			TVs: (000, 78, 4, 2, 3, 2, 195);
			Res: lpFastExtreme
		),
		(
			VGs: [lvg3902_391, lvg3931_3903up]; // 3.90.2, 3.90.3, 3.91
			TVs: (000, 78, 4, 2, 3, 2, 196);
			Res: lpFastExtreme
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (000, 78, 3, 2, 3, 4, 190);
			Res: lpStandard
		),
		(
			VGs: [lvg3931_3903up]; // 3.90.3
			TVs: (000, 78, 3, 1, 3, 4, 190);
			Res: lpStandard
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391, lvg3931_3903up]; // all relevant
			TVs: (000, 78, 4, 2, 3, 4, 190);
			Res: lpFastStandard
		),
		(
			VGs: [lvg390_3901_392]; // 3.90, 3.90.1, 3.92
			TVs: (000, 88, 4, 1, 3, 3, 195);
			Res: lpR3mix
		),
		(
			VGs: [lvg3902_391, lvg3931_3903up]; // 3.90.2, 3.90.3, 3.91
			TVs: (000, 88, 4, 1, 3, 3, 196);
			Res: lpR3mix
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (255, 99, 1, 1, 1, 2, 000);
			Res: lpStudio
		),
		(
			VGs: [lvg3931_3903up]; // 3.90.3, 3.93.1
			TVs: (255, 58, 2, 1, 3, 2, 206);
			Res: lpStudio
		),
		(
			VGs: [lvg393]; // 3.93
			TVs: (255, 58, 2, 1, 3, 2, 205);
			Res: lpStudio
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (255, 57, 2, 1, 3, 4, 205);
			Res: lpStudio
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (192, 88, 1, 1, 1, 2, 000);
			Res: lpCD
		),
		(
			VGs: [lvg3931_3903up]; // 3.90.3, 3.93.1
			TVs: (192, 58, 2, 2, 3, 2, 196);
			Res: lpCD
		),
		(
			VGs: [lvg393]; // 3.93 // 3.93
			TVs: (192, 58, 2, 2, 3, 2, 195);
			Res: lpCD
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (192, 57, 2, 1, 3, 4, 195);
			Res: lpCD
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (160, 78, 1, 1, 3, 2, 180);
			Res: lpHifi
		),
		(
			VGs: [lvg393, lvg3931_3903up]; // 3.90.3, 3.93, 3.93.1
			TVs: (160, 58, 2, 2, 3, 2, 180);
			Res: lpHifi
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (160, 57, 2, 1, 3, 4, 180);
			Res: lpHifi
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (128, 67, 1, 1, 3, 2, 180);
			Res: lpTape
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (128, 67, 1, 1, 3, 2, 150);
			Res: lpRadio
		),
		(
			VGs: [lvg390_3901_392, lvg3902_391]; // 3.90, 3.90.1, 3.90.2, 3.91, 3.92
			TVs: (112, 67, 1, 1, 3, 2, 150);
			Res: lpFM
		),
		(
			VGs: [lvg393, lvg3931_3903up]; // 3.90.3, 3.93, 3.93.1
			TVs: (112, 58, 2, 2, 3, 2, 160);
			Res: lpTapeRadioFM
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (112, 57, 2, 1, 3, 4, 160);
			Res: lpTapeRadioFM
		),
		(
			VGs: [lvg393, lvg3931_3903up]; // 3.90.3, 3.93, 3.93.1
			TVs: (056, 58, 2, 2, 0, 2, 100);
			Res: lpVoice
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (056, 57, 2, 1, 0, 4, 150);
			Res: lpVoice
		),
		(
			VGs: [lvg390_3901_392]; // 3.90, 3.90.1, 3.92
			TVs: (040, 65, 1, 1, 0, 2, 075);
			Res: lpMWUS
		),
		(
			VGs: [lvg3902_391]; // 3.90.2, 3.91
			TVs: (040, 65, 1, 1, 0, 2, 076);
			Res: lpMWUS
		),
		(
			VGs: [lvg393, lvg3931_3903up]; // 3.90.3, 3.93, 3.93.1
			TVs: (040, 58, 2, 2, 0, 2, 070);
			Res: lpMWUS
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (040, 57, 2, 1, 0, 4, 105);
			Res: lpMWUS
		),
		(
			VGs: [lvg3931_3903up]; // 3.90.3, 3.93.1
			TVs: (024, 58, 2, 2, 0, 2, 040);
			Res: lpMWEU
		),
		(
			VGs: [lvg393]; // 3.93
			TVs: (024, 58, 2, 2, 0, 2, 039);
			Res: lpMWEU
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (024, 57, 2, 1, 0, 4, 059);
			Res: lpMWEU
		),
		(
			VGs: [lvg3931_3903up]; // 3.90.3, 3.93.1
			TVs: (016, 58, 2, 2, 0, 2, 038);
			Res: lpPhone
		),
		(
			VGs: [lvg393]; // 3.93
			TVs: (016, 58, 2, 2, 0, 2, 037);
			Res: lpPhone
		),
		(
			VGs: [lvg394up]; // 3.94, 3.95
			TVs: (016, 57, 2, 1, 0, 4, 056);
			Res: lpPhone
		)
	);

function GuessPreset(const AVersionString: String; const ABitrate, AQuality,
	AEncodingMethod, ANoiseShaping, AStereoMode, AATHType, ALowpassDiv100: Byte):
	TLamePreset;

	function GuessForVersion(const AVersionGroup: TLameVersionGroup):
		TLamePreset;
	var
		LoopCounter: Integer;
	begin
		Result := lpUnknown;
		for LoopCounter := Low(PresetGuessTable) to High(PresetGuessTable) do
			if
				(AVersionGroup in PresetGuessTable[LoopCounter].VGs) and
				(PresetGuessTable[LoopCounter].TVs[0] = ABitrate) and
				(PresetGuessTable[LoopCounter].TVs[1] = AQuality) and
				(PresetGuessTable[LoopCounter].TVs[2] = AEncodingMethod) and
				(PresetGuessTable[LoopCounter].TVs[3] = ANoiseShaping) and
				(PresetGuessTable[LoopCounter].TVs[4] = AStereoMode) and
				(PresetGuessTable[LoopCounter].TVs[5] = AATHType) and
				(PresetGuessTable[LoopCounter].TVs[6] = ALowpassDiv100) then
			begin
				Result := PresetGuessTable[LoopCounter].Res;
				Break;
			end;
	end;

var
	VersionString4: String;
begin
	VersionString4 := Copy(AVersionString, 1, 4);
	{
		Both 3.90 and 3.90.1 record the version string as '3.90', whereas later
		varieties of 3.90 from 3.90.2 on record it as '3.90.'
	}
	if ((VersionString4 = '3.90') and (AVersionString <> '3.90.')) or (
			VersionString4 = '3.92') then
		Result := GuessForVersion(lvg390_3901_392)
	else if AVersionString = '3.90.' then
	begin
		{
			Both 3.90.2 and 3.90.3 record the version string as '3.90.' so need to
			test both.
		}
		Result := GuessForVersion(lvg3902_391);
		if Result = lpUnknown then
			Result := GuessForVersion(lvg3931_3903up);
	end
	else if VersionString4 = '3.91' then
		Result := GuessForVersion(lvg3902_391)
	else if VersionString4 = '3.93' then
	begin
		{
			Both 3.93 and 3.93.1 record the version string as '3.93' so need to
			test both.
		}
		Result := GuessForVersion(lvg3931_3903up);
		if Result = lpUnknown then
			Result := GuessForVersion(lvg393);
	end
	else if VersionString4 >= '3.94' then
		{
			3.94 and 3.95 are identical for preset guessing.
			3.95 is the latest version at the time or writing.
		}
		Result := GuessForVersion(lvg394up)
	else
		Result := lpUnknown;
end;

end.
